import torch
import torch.nn as nn


class Flatten(nn.Module):
    def forward(self, x, batch=True):
        if batch:
            return x.reshape(x.shape[0], -1)
        else:
            return torch.flatten(x)



import numpy as np

import torch.nn as nn

from serotiny.networks import BasicCNN


class ImageDecoder(nn.Module):
    def __init__(
        self,
        compressed_img_shape,
        hidden_channels,
        intermediate_sizes,
        latent_dim,
        output_dims,
        output_channels,
        mode,
        non_linearity,
        skip_connections,
        batch_norm,
        kernel_size,
    ):
        super().__init__()

        self.compressed_img_shape = compressed_img_shape
        compressed_img_size = np.prod(compressed_img_shape) * hidden_channels[0]
        orig_img_size = np.prod(output_dims)

        hidden_channels[-1] = output_channels
        self.hidden_channels = hidden_channels
        self.linear_decompress = nn.Linear(latent_dim, compressed_img_size)

        self.deconv = BasicCNN(
            hidden_channels[0],
            output_dim=orig_img_size,
            hidden_channels=hidden_channels,
            input_dims=compressed_img_shape,
            upsample_layers={
                i: tuple(size) for (i, size) in enumerate(intermediate_sizes)
            },
            up_conv=True,
            flat_output=False,
            kernel_size=kernel_size,
            mode=mode,
            non_linearity=non_linearity,
            skip_connections=skip_connections,
            batch_norm=batch_norm,
        )

    def forward(self, z):
        z = self.linear_decompress(z)
        z = z.view(
            z.shape[0],  # batch size
            self.hidden_channels[0],
            *self.compressed_img_shape
        )

        z = self.deconv(z)
        z = z.clamp(max=50)

        return z



import logging
import warnings
import sys

from omegaconf import OmegaConf
from hydra.utils import get_original_cwd


# silence aicsimageio related warnings
warnings.filterwarnings(action="ignore", category=FutureWarning, module="ome_types")
logging.getLogger("xmlschema").setLevel(logging.ERROR)
logging.getLogger("bfio").setLevel(logging.ERROR)
logging.getLogger("bfio.backends").setLevel(logging.ERROR)
logging.getLogger("ome_zarr").setLevel(logging.ERROR)
logging.getLogger("ome_zarr.reader").setLevel(logging.ERROR)

logger = logging.getLogger(__name__)


def instantiate(cfg):
    from hydra.utils import instantiate as _instantiate
    from copy import copy

    if not isinstance(cfg, (list, dict)):
        _cfg = OmegaConf.to_container(cfg, resolve=True)
    else:
        _cfg = copy(cfg)

    if "_aux_" in _cfg:
        del _cfg["_aux_"]

    return _instantiate(_cfg)


def _do_model_op(
    mode,
    model,
    data,
    trainer=None,
    seed=42,
    mlflow=None,
    full_conf={},
    test=False,
    predict=False,
    tune=False,
    multiprocessing_strategy=None,
    make_notebook=None,
    **_,
):

    from pathlib import Path

    import pytorch_lightning as pl

    from .mlflow_utils import mlflow_fit, mlflow_predict, mlflow_test
    from .utils import make_notebook as mk_notebook
    from .utils import save_model_predictions

    if multiprocessing_strategy is not None:
        import torch

        torch.multiprocessing.set_sharing_strategy(multiprocessing_strategy)

    if mode in ["train", "test", "predict"]:
        if make_notebook is not None:
            mk_notebook(full_conf, make_notebook)
            return

        pl.seed_everything(seed)
        logger.info("Instantiating datamodule")
        data = instantiate(data)
        logger.info("Instantiating trainer")
        trainer = instantiate(trainer)
        logger.info("Instantiating model")
        model = instantiate(model)

        if mode == "train":
            if mlflow is not None and mlflow.get("tracking_uri") is not None:
                mlflow_fit(
                    mlflow,
                    trainer,
                    model,
                    data,
                    full_conf=full_conf,
                    test=test,
                    predict=predict,
                    tune=tune,
                )
            else:
                if tune:
                    logger.info("Calling trainer.tune")
                    trainer.tune(model, data)

                logger.info("Calling trainer.fit")
                trainer.fit(model, data)

        elif mode == "test":
            if mlflow is not None and mlflow.get("tracking_uri") is not None:
                mlflow_test(mlflow, trainer, data, full_conf=full_conf)
            else:
                if "ckpt_path" not in full_conf:
                    raise NotImplementedError(
                        "Cannot `serotiny test` without "
                        "an MLFlow config, or a local ckpt_path."
                    )
                ckpt_path = full_conf["ckpt_path"]
                trainer.test(model, data, ckpt_path=ckpt_path)

        elif mode == "predict":
            if mlflow is not None and mlflow.get("tracking_uri") is not None:
                mlflow_predict(mlflow, trainer, data, full_conf=full_conf)
            else:
                if "ckpt_path" not in full_conf:
                    raise NotImplementedError(
                        "Cannot `serotiny predict` without "
                        "an MLFlow config, or a local ckpt_path."
                    )

                preds_dir_default = Path(get_original_cwd()) / "predictions"
                preds_dir = Path(
                    full_conf.get("predictions_output_dir", preds_dir_default)
                )
                preds_dir.mkdir(exist_ok=True, parents=True)

                ckpt_path = full_conf["ckpt_path"]
                preds = trainer.predict(model, data, ckpt_path=ckpt_path)
                save_model_predictions(model, preds, preds_dir)

    else:
        raise ValueError(f"`mode` must be 'train', 'test' or 'predict'. Got '{mode}'")


def _do_model_op_wrapper(cfg):
    if isinstance(cfg, dict):
        cfg = OmegaConf.create(cfg)

    # there might be other dots in the
    # executable path, ours is the last
    mode = sys.argv[0].split(".")[-1]
    if mode in ["train", "predict", "test"]:
        cfg = OmegaConf.merge(cfg, {"mode": mode})

    _do_model_op(
        **cfg,
        full_conf=cfg,
    )



import numpy as np

import torch.nn as nn
import torch


class ImageDecoder3D(nn.Module):
    def __init__(
        self, encoder, input_dims, in_channels, latent_dim,
    ):
        super().__init__()
        inp = torch.zeros(1, in_channels, *input_dims)
        compressed_img_shape = encoder(inp).shape
        self.compressed_img_shape = compressed_img_shape[2:]

        orig_img_size = np.prod(input_dims)
        deconv_layers = []
        encoder = encoder[0]
        for index in range(len(encoder) - 1, -1, -1):
            conv = encoder[index][0]
            if index == len(encoder) - 1:
                self.first_out_channels = conv.out_channels
                compressed_img_size = (
                    np.prod(self.compressed_img_shape) * self.first_out_channels
                )
            if index == 0:
                out_channels = in_channels
            else:
                out_channels = conv.out_channels

            deconv_layers.append(
                torch.nn.LazyConvTranspose3d(
                    out_channels=out_channels,
                    kernel_size=conv.kernel_size,
                    stride=conv.stride,
                )
            )
            deconv_layers.append(torch.nn.LeakyReLU())
            deconv_layers.append(torch.nn.LazyBatchNorm3d())

        deconv = nn.Sequential(*deconv_layers)
        self.deconv = deconv
        self.linear_decompress = nn.Linear(latent_dim, compressed_img_size)

    def forward(self, z):
        z = self.linear_decompress(z)
        z = z.view(
            z.shape[0],  # batch size
            self.first_out_channels,
            *self.compressed_img_shape
        )

        z = self.deconv(z)
        z = z.clamp(max=50)
        return z



from .convolution_block import ConvBlock, conv_block
from .spatial_pyramid_pool import SpatialPyramidPool, spatial_pyramid_pool
from .multiply_constant import MultiplyConstant
from .skip_connection import SkipConnection
from .flatten import Flatten
